------------------------------------------------------------------------
-- Auras
------------------------------------------------------------------------

local ADDON_NAME, NPG = ...

local min = _G.math.min
local wipe = _G.wipe
local tinsert = _G.tinsert
local tremove = _G.tremove
local ipairs = _G.ipairs
local pairs = _G.pairs
local select = _G.select
local WorldFrame = _G.WorldFrame
local UnitGUID = _G.UnitGUID
local GetTime = _G.GetTime

NPG.Auras = {}
do
   local STUPID_AURAS = { -- Auras that do something stupid, like not triggering SPELL_AURA_REFRESH when refreshing.
      [980] = 1, -- Agony
   }
   local GUIDAuras = {}
   local SpellDurations = {}
   local TrackedAuras = {}
   local CheckStupidAuras = false -- notify an OnUpdate handler to check aura durations on mouseover and target on next frame.
   local DEFAULT_SPELL_DURATION = 10
   local f = CreateFrame("Frame")
   
   
   local function GetExpirationTime(guid, spellId, spellName, isRefresh)
      local unit = (UnitGUID("mouseover") == guid and "mouseover")
         or (UnitGUID("target") == guid and "target")
      
      if unit then
         local _, _, _, _, _, duration, expiration = UnitDebuff(unit, spellName, nil, "PLAYER")
         if not SpellDurations[spellId] and not isRefresh then
            SpellDurations[spellId] = duration
         end
         
         return expiration, true
      end
      
      -- guess the expiration if spell was cast on a unit for whom we do not have an unitid (e.g. target/mouseover changed before spell finished casting)
      local remainder = 0
      if isRefresh then
         local auras = GUIDAuras[guid]
         if auras and auras[spellId] then
            remainder = auras[spellId] - GetTime()
            remainder = min(remainder, (SpellDurations[spellId] or DEFAULT_SPELL_DURATION) * 0.3)
         end
      end
      return remainder + GetTime() + (SpellDurations[spellId] or DEFAULT_SPELL_DURATION), false 
   end
   
   local function AddAura(guid, spellId, spellName, isRefresh)      
      local auras = GUIDAuras[guid]
      if not auras then
         auras = {}
         GUIDAuras[guid] = auras
      end
      
      if isRefresh and not auras[spellId] then return end
      
      local expiration, certain = GetExpirationTime(guid, spellId, spellName, isRefresh)
      auras[spellId] = expiration
      
      if not certain then
         GUIDAuras[guid].uncertain = 1
      end
   end
  
   local function RemoveAura(guid, spellId)
      local auras = GUIDAuras[guid] or {}
      if auras then auras[spellId] = nil end
   end
   
   local function CheckGUID(unit, force)
      local auras = GUIDAuras[UnitGUID(unit)]
      if auras and (auras.uncertain or force) then
         auras.uncertain = nil
         for spellId in pairs(auras) do
            local spellName = GetSpellInfo(spellId)
            local _, _, _, _, _, _, expiration = UnitDebuff(unit, spellName, nil, "PLAYER")
            auras[spellId] = expiration
         end
      end
   end
   
   function f:UPDATE_MOUSEOVER_UNIT()
      CheckGUID("mouseover", 1)
   end
   
   function f:PLAYER_TARGET_CHANGED()
      CheckGUID("target", 1)
   end
   
   function f:UNIT_TARGET(unit)
      CheckGUID(unit .. "target")
   end

   function f:COMBAT_LOG_EVENT_UNFILTERED(timestamp, event, hideCaster, sourceGUID, 
         sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, spellId, spellName)
      
      if sourceGUID == UnitGUID("player") and TrackedAuras[spellId] then
         if event == "SPELL_AURA_APPLIED" then
            AddAura(destGUID, spellId, spellName)
         elseif event == "SPELL_AURA_REFRESH" then
            AddAura(destGUID, spellId, spellName, true)
         elseif event == "SPELL_AURA_REMOVED" then
            RemoveAura(destGUID, spellId)
         elseif event == "SPELL_CAST_SUCCESS" and STUPID_AURAS[spellId] then
            CheckStupidAuras = true
         end
      end
   end
   
   
   local auraReturn = {}
   function NPG.Auras:GetAuraInfo(GUID)
      local auras = GUIDAuras[GUID]
      
      for spellId, id in pairs(TrackedAuras) do
         local aura = auraReturn[id]
         if type(aura) ~= "table" then aura = {} auraReturn[id] = aura end
         aura.spellId = spellId
         local expiration = auras and auras[spellId] or 0
         aura.remaining = expiration ~= 0 and (expiration - GetTime()) or 0
         aura.baseDuration = SpellDurations[spellId] or DEFAULT_SPELL_DURATION
      end
      
      return auraReturn
   end
   
   local TrackLater = {}
   
   function f:PLAYER_ENTERING_WORLD()
      for k,v in pairs(TrackLater) do
         local spellId = select(7,GetSpellInfo(k))
         if spellId then TrackedAuras[spellId] = v end
      end
      wipe(TrackLater)
   end
   
   local function StupidAuraCheck(f)
      if CheckStupidAuras then
         CheckStupidAuras = false
         CheckGUID("mouseover", 1)
         CheckGUID("target", 1)
      end  
   end
   
   f:SetScript("OnUpdate", StupidAuraCheck)
   
   function NPG.Auras:RegisterAuras(...)
      local numAuras = select('#', ...)
      
      f:UnregisterAllEvents()
      wipe(TrackedAuras)
      wipe(GUIDAuras)
      wipe(auraReturn)
      
      if numAuras > 0 then -- only enable aura tracking if, uh, there are any to track.
         NPG._addon.RegisterEvents(f)
         
         for i = 1, numAuras do
            local spellIdOrName = select(i, ...)
            local spellId = tonumber(spellIdOrName) or select(7,GetSpellInfo(spellIdOrName))
            if spellId then 
               TrackedAuras[spellId] = i 
            else
               TrackLater[spellIdOrName] = i
            end
         end
      end
      
   end
   
end