------------------------------------------------------------------------
-- GUID
------------------------------------------------------------------------

local ADDON_NAME, NPG = ...

local min = _G.math.min
local wipe = _G.wipe
local tinsert = _G.tinsert
local tremove = _G.tremove
local ipairs = _G.ipairs
local pairs = _G.pairs
local WorldFrame = _G.WorldFrame
local UnitGUID = _G.UnitGUID

NPG.GUID = {}
do
   local PlateGUID = {}
   local GUIDPlate = {}
   
   -- confidence values:
   -- 0: GUID not known
   -- 1: GUID acquired by comparing health delta (i.e. combat log damage&heal tally == change to previous value of health bar)  (no longer possible, absolute health values are no longer displayed on nameplates)
   -- 2: GUID acquired by matching spell casts started on a same frame update
   -- 3: GUID acquired either by mouseovering, or through combat log event on a unit with a raid target icon.
   local GUIDConfidence = {}
   setmetatable(GUIDConfidence, {__index = function() return 0 end})
   
   local GUIDFrame = CreateFrame("Frame")
   local MonitorCastStart
   local MonitorHealthDelta
   
   function NPG.GUID:GetGUIDByPlate(plateID)
      return PlateGUID[plateID]
   end
   
   function NPG.GUID:GetPlateByGUID(GUID)
      return GUIDPlate[GUID]
   end
   
   local function SetGUID(plateID, GUID, confidence)
      if confidence > GUIDConfidence[GUID] then
         PlateGUID[plateID] = GUID
         GUIDPlate[GUID] = plateID
         GUIDConfidence[GUID] = confidence
      end
   end
   
   local function WipePlateGUID(plateID)
      local GUID = PlateGUID[plateID]
      if GUID then
         GUIDPlate[GUID] = nil
         GUIDConfidence[GUID] = nil
      end
      PlateGUID[plateID] = nil
   end
   
   do
      local f = GUIDFrame
      local PlateCastsStarted = {}
      local PlateHealthDeltas = {} 
      local GUIDCastsStarted = {}
      local GUIDHealthDeltas = {}
      local NumPlateCastsStarted = 0
      local NumPlateHealthDeltas = 0
      
      local REACTION_FRIENDLY = COMBATLOG_OBJECT_REACTION_FRIENDLY + COMBATLOG_OBJECT_REACTION_NEUTRAL
      local REACTION_HOSTILE = COMBATLOG_OBJECT_REACTION_HOSTILE + COMBATLOG_OBJECT_REACTION_NEUTRAL
      
      local RAIDTARGET_MASK = COMBATLOG_OBJECT_RAIDTARGET_MASK
      local RAIDTARGETS = 
         { [COMBATLOG_OBJECT_RAIDTARGET1] = 1, [COMBATLOG_OBJECT_RAIDTARGET2] = 2,
           [COMBATLOG_OBJECT_RAIDTARGET3] = 3, [COMBATLOG_OBJECT_RAIDTARGET4] = 4,
           [COMBATLOG_OBJECT_RAIDTARGET5] = 5, [COMBATLOG_OBJECT_RAIDTARGET6] = 6,
           [COMBATLOG_OBJECT_RAIDTARGET7] = 7, [COMBATLOG_OBJECT_RAIDTARGET8] = 8 }
         
      local band = bit.band
      
      local function NoDuplicateInsert(table, key, value)
         local existingValue = table[key]
         
         if existingValue == nil then
            table[key] = value
            return 1
         elseif existingValue ~= false then
            table[key] = false
            return -1
         end
         
         return 0
      end
      
      local function CheckRaidTarget(GUID, flag)
         if GUIDConfidence[GUID] < 3 then
            local iconID = RAIDTARGETS[flag]
            local plateID = NPG.Plates:GetPlateByRaidIcon(iconID)
            if plateID then
               SetGUID(plateID, GUID, 3)
            end
         end
      end
      
      function f:UPDATE_MOUSEOVER_UNIT()
         if GetMouseFocus() == WorldFrame then
            local GUID = UnitGUID("mouseover")
            
            if GUID and GUIDConfidence[GUID] < 3 then
               local plateID = NPG.Plates:GetHighlightedPlate()
               if plateID then
                  SetGUID(plateID, GUID, 3)
               end
            end
         end
      end
      
      f.CURSOR_UPDATE = f.UPDATE_MOUSEOVER_UNIT
      
      function f:COMBAT_LOG_EVENT_UNFILTERED(timestamp, event, hideCaster, sourceGUID, 
            sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, ...)
         
         local canBeSource = false
         local canBeDest = false
         
         local friendPlates = tonumber(GetCVar("nameplateShowFriends")) == 1
         local enemyPlates = tonumber(GetCVar("nameplateShowEnemies")) == 1
         
         if enemyPlates then
            if band(sourceFlags, REACTION_HOSTILE) > 0 then 
               canBeSource = true
            end
            if band(destFlags, REACTION_HOSTILE) > 0 then 
               canBeDest = true
            end
         end
         
         if friendPlates then
            if band(sourceFlags, REACTION_FRIENDLY) > 0 then 
               canBeSource = true
            end
            if band(destFlags, REACTION_FRIENDLY) > 0 then
               canBeDest = true
            end
         end
         
         if canBeDest and (band(destRaidFlags, RAIDTARGET_MASK) > 0) then
            CheckRaidTarget(destGUID, destRaidFlags)
         end
         
         if canBeSource and (band(sourceRaidFlags, RAIDTARGET_MASK) > 0) then
            CheckRaidTarget(sourceGUID, sourceRaidFlags)
         end
         
         if event == "SPELL_CAST_START" then
            if canBeSource and GUIDConfidence[sourceGUID] < 2 then
               local _, spellName = ...
               
               NoDuplicateInsert(GUIDCastsStarted, spellName, sourceGUID)
            end
            
            return
         end
         
         if not canBeDest or GUIDConfidence[destGUID] > 1 then return end
         
         local currentHealth = GUIDHealthDeltas[destGUID] or 0
         
         if event == "ENVIRONMENTAL_DAMAGE" then
            local _, damage = ...
            GUIDHealthDeltas[destGUID] = currentHealth - damage
            
         elseif event == "SWING_DAMAGE" then
            local damage = ...
            GUIDHealthDeltas[destGUID] = currentHealth - damage
            
         elseif event:sub(-7) == "_DAMAGE" then
            local _, _, _, damage = ...
            GUIDHealthDeltas[destGUID] = currentHealth - damage
         elseif event:sub(-5) == "_HEAL" then
            local _, _, _, heal = ...
            GUIDHealthDeltas[destGUID] = currentHealth + heal
         end
         
      end
      
      MonitorCastStart = function(spellName, plateID)
         NumPlateCastsStarted = NumPlateCastsStarted + NoDuplicateInsert(PlateCastsStarted, spellName, plateID)
      end
      
      MonitorHealthDelta = function(delta, plateID)
         NumPlateHealthDeltas = NumPlateHealthDeltas + NoDuplicateInsert(PlateHealthDeltas, delta, plateID)
      end
      
      local function CheckCastsStarted()
         if NumPlateCastsStarted ~= 0 then
            for spell, plateID in pairs(PlateCastsStarted) do
               local GUID = GUIDCastsStarted[spell]
               if GUID and plateID then
                  SetGUID(plateID, GUID, 2)
               end
            end
         end
         
         wipe(PlateCastsStarted)
         wipe(GUIDCastsStarted)
         NumPlateCastsStarted = 0
      end
      
      local tmpGUIDDeltas = {}
      local function CheckHealthDeltas()
         if NumPlateHealthDeltas ~= 0 then
            wipe(tmpGUIDDeltas)
            
            -- convert GUID->delta to delta->GUID
            for GUID, delta in pairs(GUIDHealthDeltas) do
               NoDuplicateInsert(tmpGUIDDeltas, delta, GUID)
            end
            
            for delta, GUID in pairs(tmpGUIDDeltas) do
               local plateID = PlateHealthDeltas[delta]
               if GUID and plateID then
                  SetGUID(plateID, GUID, 1)
               end
            end
            
            
            wipe(PlateHealthDeltas)
            wipe(GUIDHealthDeltas)
            NumPlateHealthDeltas = 0
         end
      end
      
      NPG._addon.RegisterEvents(f)
      
      f:SetScript("OnUpdate", function()
         CheckCastsStarted()
         CheckHealthDeltas()
      end)
   end
   
   do
      local PlateHealth = {}
      
      local function Plate_OnShow(plateID)
         WipePlateGUID(plateID)
         PlateHealth[plateID] = nil
      end
      
      local function Plate_OnHide(plateID)
         WipePlateGUID(plateID)
      end
      
      
      local function Health_OnValue(plateID, plate, parts, value)
         if GUIDConfidence[plateID] >= 1 then return end
         
         local previousHealth = PlateHealth[plateID]
         if previousHealth then
            local delta = value - previousHealth
            MonitorHealthDelta(delta, plateID)
         end
         
         PlateHealth[plateID] = value
      end
      
      local function Cast_OnShow(plateID, plate, parts)
         MonitorCastStart(parts.castName:GetText(), plateID)
      end
   
      function NPG.GUID:EnableHooks()
         NPG.Plates:AddHook("Plate_OnShow", Plate_OnShow)
         NPG.Plates:AddHook("Plate_OnHide", Plate_OnHide)
         --NPG.Plates:AddHook("Health_OnValue", Health_OnValue)
         NPG.Plates:AddHook("Cast_OnShow", Cast_OnShow)
      end
      
      function NPG.GUID:DisableHooks()
         NPG.Plates:RemoveHook("Plate_OnShow", Plate_OnShow)
         NPG.Plates:RemoveHook("Plate_OnHide", Plate_OnShow)
         --NPG.Plates:RemoveHook("Health_OnValue", Health_OnValue)
         NPG.Plates:RemoveHook("Cast_OnShow", Cast_OnShow)
      end
   end
end