------------------------------------------------------------------------
-- Plates
------------------------------------------------------------------------

local ADDON_NAME, NPG = ...

local wipe = _G.wipe
local tinsert = _G.tinsert
local tremove = _G.tremove
local pairs = _G.pairs
local WorldFrame = _G.WorldFrame
local UnitGUID = _G.UnitGUID

NPG.Plates = {}
do
   local PlateCache = {}
   local PlatePartsCache = {}
   local VisiblePlates = {}
   
   local FOREIGN_SERVER_PATTERN = _G.FOREIGN_SERVER_LABEL:gsub("[()*]", "%%%1")
   
   local Hooks = {
      Plate_OnShow = {},
      Plate_OnHide = {},
      Health_OnValue = {},
      Cast_OnShow = {},
      Cast_OnHide = {},
      Cast_OnValue = {}
   }
   
   local function RunHooks(hookName, ...)
      local hooks = Hooks[hookName]
      if hooks then
         for i = 1, #hooks do
            hooks[i](...)
         end
      end
   end

   local function SetHooks(plateID, plate, parts)
      plate:HookScript("OnShow", function() RunHooks("Plate_OnShow", plateID, plate, parts) end)
      plate:HookScript("OnHide", function() RunHooks("Plate_OnHide", plateID, plate, parts) end)
      parts.health:HookScript("OnValueChanged", function(self, value) RunHooks("Health_OnValue", plateID, plate, parts, value) end)
      parts.cast:HookScript("OnShow", function() RunHooks("Cast_OnShow", plateID, plate, parts) end)
      parts.cast:HookScript("OnHide", function() RunHooks("Cast_OnHide", plateID, plate, parts) end)
      parts.cast:HookScript("OnValueChanged", function(self, value) RunHooks("Cast_OnValue", plateID, plate, parts, value) end)
   end
   
   function NPG.Plates:AddHook(hookType, hookFunc)
      local hooktbl = Hooks[hookType]
      
      if hooktbl and not tContains(hooktbl, hookfunc) then
         tinsert(hooktbl, hookFunc)
      end
   end
   
   function NPG.Plates:RemoveHook(hookType, hookFunc)
      local hooktbl = Hooks[hookType]
      
      if hooktbl then
         local id = tContains(hooktbl, hookfunc)
         if id then 
            tremove(hooktbl, id)
         end
      end
   end
   
   NPG.Plates:AddHook("Plate_OnShow", function(plateID, plate) VisiblePlates[plate] = plateID end)
   NPG.Plates:AddHook("Plate_OnHide", function(plateID, plate) VisiblePlates[plate] = false end)
   
   local function CachePlate(plateID)
      plate = _G["NamePlate"..plateID]
      if not plate or PlateCache[plateID] then return end
      
      PlateCache[plateID] = plate
      
      local parts = {}
      local frame, nameframe = plate:GetChildren()
      parts.frame = frame
      parts.nameframe = nameframe
      
      parts.health, parts.cast = frame:GetChildren()
      parts.name = nameframe:GetRegions()
      
      local threat, _, highlight, level, skullIcon, raidIcon, eliteIcon = frame:GetRegions()
      parts.threat = threat
      parts.highlight = highlight
      parts.level = level
      parts.skullIcon = skullIcon
      parts.raidIcon = raidIcon
      parts.eliteIcon = eliteIcon
      
      local _, _, castShield, castIcon, castName = parts.cast:GetRegions()
      parts.castShield = castShield
      parts.castIcon = castIcon
      parts.castName = castName
      
      PlatePartsCache[plateID] = parts
      
      SetHooks(plateID, plate, parts)
      
      -- manually fire OnShow since we haven't hooked yet when it fires :C
      if plate:IsVisible() then
         RunHooks("Plate_OnShow", plateID, plate, parts)
      end
      
      return plate
   end
   
   function NPG.Plates:HideID(plateID)
      local parts = self:GetPlateParts(plateID)
      
      parts.nameframe:SetAlpha(0)
      parts.frame:SetAlpha(0)
   end
   
   function NPG.Plates:ShowID(plateID)
      local parts = self:GetPlateParts(plateID)
      
      parts.nameframe:SetAlpha(1)
      parts.frame:SetAlpha(1)
   end
   
   function NPG.Plates:GetPlate(plateID)
      return PlateCache[plateID] or CachePlate(plateID)
   end
   
   function NPG.Plates:GetPlateParts(plateID)
      local parts = PlatePartsCache[plateID]
      if not parts then CachePlate(plateID) end
      
      return PlatePartsCache[plateID]
   end
   
   function NPG.Plates:GetPlateData(plateID, data)
      local parts = self:GetPlateParts(plateID)
      if not parts then return end
   
      data.minhealth, data.maxhealth = parts.health:GetMinMaxValues() -- minhealth is probably always 0. probably!
      data.health = parts.health:GetValue()
      data.name = parts.name:GetText():gsub(FOREIGN_SERVER_PATTERN, "")
      data.icon = self:GetRaidIconByPlate(plateID)
      data.isElite = parts.eliteIcon:IsShown()
      data.isBoss = parts.skullIcon:IsShown()
      data.GUID = NPG.GUID:GetGUIDByPlate(plateID)
      data.aura = NPG.Auras:GetAuraInfo(data.GUID)
      data.level = parts.level:GetText()
      data.casting = parts.cast:IsShown() and (parts.castShield:IsShown() and 2 or 1)
      
      return parts
   end
  
   ------------------------------------------------------------------------------
   -- Stuff that must be checked manually on any new frame update.
   -- We'll only check it on demand and cache the data for the remainder of the frame.
   do
      local PlateRaidIcon = {}
      local RaidIconPlate = {}
      local HighlightedPlate = nil
      local DataIsFresh = false
      
      local RAIDICON_COORDINATES = {
         [0]    = {[0] = 1, [0.25] = 5},
         [0.25] = {[0] = 2, [0.25] = 6},
         [0.5]  = {[0] = 3, [0.25] = 7},
         [0.75] = {[0] = 4, [0.25] = 8}
      }
      
      local function UpdateData()
         if DataIsFresh then return end
         
         wipe(PlateRaidIcon)
         wipe(RaidIconPlate)
         HighlightedPlate = nil
         
         for plate, plateID in pairs(VisiblePlates) do
            local parts = PlatePartsCache[plateID]
            if parts then
               if parts.highlight:IsShown() then
                  HighlightedPlate = plateID
               end
             
               if parts.raidIcon:IsShown() then
                  local ULx, ULy = parts.raidIcon:GetTexCoord()
                  local iconID = RAIDICON_COORDINATES[ULx]
                  iconID = iconID and iconID[ULy]
                  
                  if iconID then
                     PlateRaidIcon[plateID] = iconID
                     RaidIconPlate[iconID] = plateID
                  end
               end
            end
         end
         
         DataIsFresh = 1
      end
      
      function NPG.Plates:GetPlateByRaidIcon(raidIcon)
         UpdateData()
         return RaidIconPlate[raidIcon]
      end
      
      function NPG.Plates:GetRaidIconByPlate(plateID)
         UpdateData()
         return PlateRaidIcon[plateID]
      end
      
      function NPG.Plates:GetHighlightedPlate()
         UpdateData()
         return HighlightedPlate
      end
      
      local Updater = CreateFrame("Frame")
      Updater:SetScript("OnUpdate", function() DataIsFresh = false end)
   end
      
end